/*
 *    Structure Class
 *
 */

import java.io.*;
import java.lang.*;
import java.util.*;


/**
 * Structure (Stem List + Sequence)
 *
 */
public class Structure {
	private ArrayList stemList;
	private Sequence sequence;


	/**
	 * Create a new Structure
	 */
	public Structure() {
		stemList = new ArrayList();
	}

	/**
	 * Print the Structure in text mode (for debugging)
	 */
	public void printStemList() {
		System.out.println(sequence.getName());
		System.out.println(sequence.getStr());
		for (int i=0; i < stemList.size(); i++) {
			Stem stem = (Stem) stemList.get(i);
			System.out.println ("[" + stem.i1 + "," + stem.i2 + ":" + stem.i3 + "," + stem.i4 + "]");
		}
	}


	/**
	 * get a Array list of the Stems of the Structure
	 */
	public ArrayList getStemList() {
		return stemList;
	}
	
	/**
	 * get the Sequence of the Structure
	 */
	public Sequence getSequence() {
		return sequence;
	}

	/**
	 * Parse a ct File (return -1 if error) and construct the Structure associated
	 */
	public int parse(File file) {
		try {
			if (parseCt(file) == -1)
				return -1;
		} catch (IOException e) {System.err.println("ERROR while parsing File:" + file.getName()); return -1;}
		catch (NoSuchElementException n) {System.err.println("ERROR while parsing File:" + file.getName()); return -1;}
		return 1;
	}
	private int parseCt(File file) throws IOException{
		BufferedReader buffer;
		try {
			buffer = new BufferedReader(new FileReader(file));
		}catch (FileNotFoundException e) {
			System.err.println("ERROR: can't open file: '" + file.getName() + "'");
                return -1;
		}
		
		String line;
		StringTokenizer tokenizer;
		StringBuffer sequenceBuffer = new StringBuffer();
		String sequenceName;

		boolean started = false;
		if ((line = buffer.readLine()) == null)
			return -1;
		tokenizer = new StringTokenizer(line);
		tokenizer.nextToken();
		sequenceName = tokenizer.nextToken();
		
		
		int i,j; /* abciss of the open and close part of a stem */
		i = 0; j = 0;
		int lastJ = 0;
		int i1, i2, i3, i4;
		i1 = 0;
		i2 = 0;
		i3 = 0;
		i4 = 0;
		
		while (tokenizer.hasMoreElements())
			sequenceName = sequenceName + " " + tokenizer.nextToken();
		
		while ((line = buffer.readLine()) != null) {
			if (line.compareTo("") == 0)
				return -1;
			tokenizer = new StringTokenizer(line);
			
			i = Integer.parseInt(tokenizer.nextToken());
			sequenceBuffer.append(tokenizer.nextToken().toLowerCase());
			
			tokenizer.nextToken();
			tokenizer.nextToken();
			
			lastJ = j;
			j = Integer.parseInt(tokenizer.nextToken());
			
			if (!started) {
				if (j != 0 && i < j){ /* start stem */
					i1 = i;
					i4 = j;
					started = true;
				}
			}else{
				if (j != lastJ - 1) { /* stop stem */
					i2 = i - 1;
					i3 = lastJ;
					stemList.add(new Stem(i1-1, i2-1, i3-1, i4-1));
					started = false;
				}
			}
		}
		
		//sequence = new Sequence(sequenceBuffer.toString(), sequenceName, Sequence.getFirst(file));
		sequence = new Sequence(sequenceBuffer.toString(), sequenceName, 0);
		return 1;
	
	}
/*
	public static void main(String[] args) {
		Structure struct = new Structure();
		struct.parse(new File("str.001.002.ct"));
		struct.printStemList();
	
	}
*/
}
