/*
 * CARNAC interface
 * prediction of secondary structure of RNA
 *
 */

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.net.*;

/**
 * This is the Structure Frame
 * 
 * graphical frame with structure backbone representation
 */
public class StructureFrame extends JFrame{


	/*
	 * Swing relative
	 */
	private ZoomBar zoomBar;
	private JMenuBar menuBar;
	private static final String iconsPath = new String("resources/icons/");
	private int blocIncrementWidth = 40;
	
	private GraphicsPanel graphicsPanel;
	private JScrollPane scrollPane;
	private File ctFile;
	private Structure structure;
	private Sequence sequence;
	
    /**
     * Construct a new Panel for displaying The Secondary Structure
     */
	public StructureFrame(int locX, int locY) {
		super("Secondary Structure view");

		Dimension minimumSize = new Dimension(380, 120);
		Dimension size =  new Dimension(785, 240);
		//setMinimumSize(minimumSize);
		setLocation(locX, locY);
		setSize(size);
		//setLocation(0, 30);
		setJMenuBar(new MenuBar());
		getContentPane().add(new ZoomBar(), BorderLayout.WEST);
		
		
		graphicsPanel = new GraphicsPanel(this);
		structure = new Structure();
			
		setVisible(true);
		/* Relative to Scrolling */
		JScrollPane scrollPane = new JScrollPane(graphicsPanel);
		scrollPane.getHorizontalScrollBar().setBlockIncrement(blocIncrementWidth);
		scrollPane.getHorizontalScrollBar().setUnitIncrement(blocIncrementWidth);
		scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		
		getContentPane().add(scrollPane, BorderLayout.CENTER);
		
		addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });
		
	}


	public void show(File ctFile) {
			
		if (structure.parse(ctFile) == -1)
			return;
		this.ctFile = ctFile;
		sequence = structure.getSequence();
		graphicsPanel.set(structure);
		setTitle("[ " +  sequence.getName() + " ] Structure");
				
		setVisible(true);
		graphicsPanel.showAll();

		
	}

	JButton zoomInButton, zoomOutButton, zoomToFitButton, zoomMaxButton, viewSequenceButton, zoomSelectionButton;
	
	/*
	 * Actions
	 */
	public void toggleZoomSelection() {

		if (graphicsPanel.zoomSelection()) {
			zoomSelectionButton.setIcon(new ImageIcon(ZoomBar.class.getResource(iconsPath + "UnzoomSelection24.gif")));
			zoomSelectionButton.updateUI();
			zoomSelectionItem.setIcon(new ImageIcon(ZoomBar.class.getResource(iconsPath + "UnzoomSelection16.gif")));
			zoomSelectionItem.updateUI();
		}else{
			zoomSelectionButton.setIcon(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomSelection24.gif")));
			zoomSelectionButton.updateUI();
			zoomSelectionItem.setIcon(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomSelection16.gif")));
			zoomSelectionItem.updateUI();
		}
	}
	
	private class ZoomBar extends JToolBar {
		
		public ZoomBar() {
			setOrientation(JToolBar.VERTICAL);
			setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
			addSeparator(new Dimension(2, 2));

			zoomInButton = new JButton(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomIn24.gif")));
			zoomInButton.setMargin(new Insets(1,1,1,1));
			zoomInButton.addActionListener(zoomIn());
			
			zoomOutButton = new JButton(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomOut24.gif")));
			zoomOutButton.setMargin(new Insets(1,1,1,1));
			zoomOutButton.addActionListener(zoomOut());
			
			JToolBar zoomPlusMinusBar = new JToolBar(BoxLayout.X_AXIS);
			zoomPlusMinusBar.add(zoomOutButton);
			zoomPlusMinusBar.add(zoomInButton);
			zoomPlusMinusBar.setAlignmentX(Component.CENTER_ALIGNMENT);
			zoomPlusMinusBar.setFloatable(false);
			add(zoomPlusMinusBar);
			
			addSeparator(new Dimension(2, 2));

			zoomToFitButton = new JButton(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomFit24.gif")));
			zoomToFitButton.setMargin(new Insets(1,1,1,1));
			zoomToFitButton.addActionListener(zoomToFit());
			zoomToFitButton.setAlignmentX(Component.CENTER_ALIGNMENT);
			add(zoomToFitButton);

			zoomMaxButton = new JButton(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomMax24.gif")));
			zoomMaxButton.setMargin(new Insets(1,1,1,1));
			zoomMaxButton.addActionListener(zoomMax());
			zoomMaxButton.setAlignmentX(Component.CENTER_ALIGNMENT);
			add(zoomMaxButton);

			viewSequenceButton = new JButton(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomSequence24.gif")));
			viewSequenceButton.setMargin(new Insets(1,1,1,1));
			viewSequenceButton.addActionListener(viewSequence());
			viewSequenceButton.setAlignmentX(Component.CENTER_ALIGNMENT);
			add(viewSequenceButton);
			
			zoomSelectionButton = new JButton(new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomSelection24.gif")));
			zoomSelectionButton.setMargin(new Insets(1,1,1,1));
			zoomSelectionButton.addActionListener(zoomSelection());
			zoomSelectionButton.setAlignmentX(Component.CENTER_ALIGNMENT);
			add(zoomSelectionButton);			

			setFloatable(false);

		}
	}
	
	private void checkEnabled() {
		zoomMaxButton.setEnabled(true);
		zoomInButton.setEnabled(true);
		zoomOutButton.setEnabled(true);
		
		zoomInItem.setEnabled(true);
		zoomOutItem.setEnabled(true);
		zoomMaxItem.setEnabled(true);
	
		if (graphicsPanel.isZoomMin()){
			zoomOutButton.setEnabled(false);
			zoomOutItem.setEnabled(false);
	
		}
		if (graphicsPanel.isZoomMax()) {
			zoomInButton.setEnabled(false);
			zoomMaxButton.setEnabled(false);
			zoomInItem.setEnabled(false);
			zoomMaxItem.setEnabled(false);
		}
	}

	/*
	 * Actions 
	 */
	protected  ActionListener zoomToFit() {
		return new ActionListener() {
			public void actionPerformed(ActionEvent e)  {
				graphicsPanel.showAll();
				checkEnabled();
			}
		};
	}
	protected  ActionListener zoomMax() {
		return new ActionListener() {
			public void actionPerformed(ActionEvent e)  {
				graphicsPanel.zoomMax();
				checkEnabled();
			}
		};
	}
/*	
	protected  ActionListener zoomMin() {
		return new ActionListener() {
			public void actionPerformed(ActionEvent e)  {
				graphicsPanel.zoomMin();
				checkEnabled();
			}
		};
	}
*/
	protected  ActionListener zoomSelection() {
		return new ActionListener() {
			public void actionPerformed(ActionEvent e)  {
				graphicsPanel.ToggleZoomToSelection();
				toggleZoomSelection();
				checkEnabled();
			}
		};
	}	
	protected  ActionListener zoomIn() {
		return new ActionListener() {
			public void actionPerformed(ActionEvent e)  {
                graphicsPanel.setScale(graphicsPanel.getScale() + 0.2f);
				checkEnabled();
			}
		};
	}
	
	protected  ActionListener zoomOut() {
		return new ActionListener() {
			public void actionPerformed(ActionEvent e)  {
                graphicsPanel.setScale(graphicsPanel.getScale() - 0.2f);
				checkEnabled();
			}
		};
	}
	
	protected  ActionListener viewSequence() {
		return new ActionListener() {
			public void actionPerformed(ActionEvent e)  {
				graphicsPanel.swapViewSequence();
				checkEnabled();
			}
		};
	}	
	
	private JMenuItem zoomMaxItem, zoomInItem, zoomOutItem, zoomToFitItem, viewSequenceItem, zoomSelectionItem;
    private JRadioButtonMenuItem font12Item, font14Item, font16Item;
    /**
     *   Menu
     */
    private class MenuBar extends JMenuBar {


		public MenuBar() {
			/*
			 * File Menu
			 */
			JMenu fileMenu = new JMenu("File");
			JMenuItem exitItem = new JMenuItem("Exit");
			fileMenu.add(exitItem);
			exitItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q , ActionEvent.ALT_MASK));
			exitItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e){
					System.exit(0);
				}
            });
		
			/*
			 * Zoom Menu
			 */
			JMenu zoomMenu = new JMenu("Zoom");
			zoomMaxItem = new JMenuItem("Zoom To Max", new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomMax16.gif")));
			zoomMenu.add(zoomMaxItem);
			zoomMaxItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_NUMPAD8 , 0));
			zoomMaxItem.addActionListener(zoomMax());
				
			zoomToFitItem = new JMenuItem("Zoom to Fit", new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomFit16.gif")));
			zoomToFitItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_NUMPAD5 , 0));
	
			zoomMenu.add(zoomToFitItem);
			zoomToFitItem.addActionListener(zoomToFit());
				
			zoomSelectionItem = new JMenuItem("Zoom to Selection", new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomSelection16.gif")));
			zoomSelectionItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DECIMAL , 0));
			zoomMenu.add(zoomSelectionItem);
			zoomSelectionItem.addActionListener(zoomSelection());
		
			zoomInItem = new JMenuItem("Zoom +", new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomIn16.gif")));
			zoomInItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_ADD , 0));
			zoomMenu.add(zoomInItem);
			zoomInItem.addActionListener(zoomIn());
				
			zoomOutItem = new JMenuItem("Zoom -", new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomOut16.gif")));
			zoomOutItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_SUBTRACT , 0));
			zoomMenu.add(zoomOutItem);
			zoomOutItem.addActionListener(zoomOut());
	
			/*
			 * View Menu
			 */	
			JMenu viewMenu = new JMenu("View");
		
			viewSequenceItem = new JMenuItem("Toggle Sequence View",  new ImageIcon(ZoomBar.class.getResource(iconsPath + "ZoomSequence16.gif")));
			viewSequenceItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_MULTIPLY , 0));
			viewMenu.add(viewSequenceItem );
			viewSequenceItem.addActionListener(viewSequence());
			viewMenu.addSeparator();

			JMenuItem naviewSequenceItem = new JMenuItem("View naview");
			naviewSequenceItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, ActionEvent.ALT_MASK));
			viewMenu.add(naviewSequenceItem );
			naviewSequenceItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e){
						NaviewThread naviewThread = new NaviewThread(ctFile);
					}
            });
	
	
			JMenuItem moveRightItem = new JMenuItem("Move Right");
			viewMenu.add(moveRightItem);
			moveRightItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_NUMPAD6 , 0));
			moveRightItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e){
					graphicsPanel.moveVisible(+ (int) ( (float) getContentPane().getWidth() / 2f));
				}
            });
			JMenuItem moveLeftItem = new JMenuItem("Move Left");
			viewMenu.add(moveLeftItem);
			moveLeftItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_NUMPAD4 , 0));
			moveLeftItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e){
					graphicsPanel.moveVisible(+ (int) (- (float) getContentPane().getWidth() / 2f));
				}
            });
		
			
			/*
			 * Selection Menu
			 */
			JMenu selectMenu = new JMenu("Selection");
	   
			JMenuItem selectNoneItem = new JMenuItem("Select None");
			selectNoneItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
			selectMenu.add(selectNoneItem);
			selectNoneItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e){
					graphicsPanel.selectNone();
				}
            });
			JMenuItem selectNextItem = new JMenuItem("Select Next");
			selectNextItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_NUMPAD3, 0));
			selectMenu.add(selectNextItem);
			selectNextItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e){
					graphicsPanel.selectNext();
				}
            });
			JMenuItem selectPreviousItem = new JMenuItem("Select Previous");
			selectPreviousItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_NUMPAD1, 0));
			selectMenu.add(selectPreviousItem);
			selectPreviousItem.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e){
					graphicsPanel.selectPrevious();
				}
            });
		 
			/*
			 * Settings Menu
			 */
			JMenu settingsMenu = new JMenu("settings");
			
			font12Item = new  JRadioButtonMenuItem("font 12");
			settingsMenu.add(font12Item);
		    font12Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e){
				graphicsPanel.changeFont( new Font("Monospaced", Font.BOLD, 12));
				font12Item.setSelected(true);
				font14Item.setSelected(false);
				font16Item.setSelected(false);
			}
            });

			font14Item = new  JRadioButtonMenuItem("font 14", true);
			settingsMenu.add(font14Item);
			font14Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e){
				graphicsPanel.changeFont(new Font("Monospaced", Font.BOLD, 14));
				font12Item.setSelected(false);
				font14Item.setSelected(true);
				font16Item.setSelected(false);
			}
            });
		 
			font16Item = new  JRadioButtonMenuItem("font 16");
			settingsMenu.add(font16Item);
			font16Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e){
				graphicsPanel.changeFont( new Font("Monospaced", Font.BOLD, 16));
				font12Item.setSelected(false);
				font14Item.setSelected(false);
				font16Item.setSelected(true);
			}
            });
		 
			add(fileMenu);
			add(selectMenu);
			add(viewMenu);
			add(zoomMenu);
			add(settingsMenu);

		}	 
	 }
	
}
