
// Fasta2rnaml converts (FASTA+CM) to RNAML

#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#ifndef SIZE_MAX_NUCL
#define	SIZE_MAX_NUCL 1000
#define	SIZE_MAX_NAME 80
#define	NB_MAX_STEMS 30
#define print(x) printf("\n %i\n\n",x);fflush(stdout)
#define print_(x) printf("%i.",x);fflush(stdout)
#define step printf("\n #\n\n");fflush(stdout)
#define step_ printf("#.");fflush(stdout)
#define nnn printf("\n")
#endif

int is (char x, char *alphabet)
{int i,j; char a;
 i=0; j=0; while ((a=alphabet[i]) != '\0')
   {if (x == alphabet[i]) j=1; i++;} return j;}

char *new_string (char *in)
{return strdup (in);}

char *cat (char *prefix, char *suffix) {
  char *aux; char *str2;
  aux = (char*) malloc((strlen (prefix)+strlen(suffix)+1)*sizeof(char));
  strcpy (aux, prefix); str2 = strdup (suffix);
  aux = strcat (aux, str2); free (prefix);
  free (str2); return aux;}

int main (int argc, char **argv) {

  char *base = "AUCGaucg";
  char **all_seq_nucl, **all_seq_name;
  char *name_out, *junk;
  int nb_seq;
  int *nb_stems, *bp;
  int i, j, k, l, m, x, y, length;
  char c;
  FILE *infa, *incm, *outrnaml;

  // usage: Fasta2rnaml <file.fasta> <file.cm>

  // read sequences in fasta input file
  // comment by a $ is disallowed

  if ((infa=fopen(argv[1],"r")) == NULL)
    {printf ("\n usage:  Fasta2rnaml <file.fasta> <file.cm>\n\n"); exit(0);}
  nb_seq = 0; while ((c=fgetc(infa))!=EOF) if (c=='>') nb_seq++;
  all_seq_nucl = (char**) malloc (nb_seq * sizeof(char*));
  all_seq_name = (char**) malloc (nb_seq * sizeof(char*));
  fclose(infa);

  infa=fopen(argv[1],"r");

  while ((c != '>') && (c != EOF)) c=fgetc(infa); c=fgetc(infa);
  // now c == '>' ; first sequence to be read

  for (j=0; j<nb_seq; j++) {
    all_seq_nucl[j] = (char*) calloc (SIZE_MAX_NUCL, sizeof(char));
    all_seq_name[j] = (char*) calloc (SIZE_MAX_NAME, sizeof(char));

    // sequence name
    while ((c != '>') && (c != EOF)) c=fgetc(infa);
    fgets((all_seq_name[j]),SIZE_MAX_NAME,infa);
    (all_seq_name[j])[strlen(all_seq_name[j])-1]='\0';
    printf("\n  [FASTA] reading : [%2i] %s", j+1, all_seq_name[j]);

    // sequence: nucl = nucl[0..n-1]
    i=0; while (((c = fgetc(infa)) != '>') && (c != EOF))
      {if (is (c,base)) (all_seq_nucl[j])[i++]=(char)toupper((int)c);
      else if ((c=='t')||(c=='T')) (all_seq_nucl[j])[i++]='u';}
    (all_seq_nucl[j])[i]='\0';
  }

  nnn;
  fclose (infa);

  // read structures in cm input file

  // allocate
  junk = (char*) calloc (SIZE_MAX_NAME, sizeof(char));
  nb_stems = (int*) calloc (nb_seq, sizeof(int));

  if ((incm=fopen(argv[2],"r")) == NULL)
    {printf ("\n\n usage:  Fasta2rnaml <file.fasta> <file.cm>\n\n"); exit(0);}
  c=fgetc(incm);
  while ((c != '>') && (c != EOF)) c=fgetc(incm);
  // now c == '>' ;

  // store nb stems by structure
  i=0;
  while ((i<nb_seq) && (c != EOF) && (c != '$')) {
    while ((c != '>') && (c != EOF)) c=fgetc(incm);
    c=fgetc(incm);
    fgets(junk, SIZE_MAX_NAME, incm);
    for (l=0; l<NB_MAX_STEMS; l++)
      fscanf (incm, "%i%*i%*i%*i", &nb_stems[i]);
    //    printf ("[%i] %i\n", i+1, nb_stems[i]);
    i++;
  }
  //  nnn; nnn;
  fclose (incm);

  //allocate bp
  bp = (int*) calloc (SIZE_MAX_NUCL, sizeof(int));

  name_out = new_string(argv[1]);
  name_out[strlen(name_out)-6]='\0';
  name_out = cat(name_out, new_string("_forced.rnaml"));

  incm=fopen(argv[2],"r");
  outrnaml = fopen(name_out,"w");

  i=-1; c=fgetc(incm);
  while ((c != '>') && (c != EOF)) c=fgetc(incm);
  // now c == '>' ; first list of stems to be read

  while ((c != EOF) && (c != '$')) {

    // init bp
    for (l=0; l<SIZE_MAX_NUCL; l++) bp[l]=0;

    // i
    while ((c != '>') && (c != EOF)) c=fgetc(incm);
    if (c=='>') i++;

    // reading stems
    c=fgetc(incm);
    if ((c != EOF) && (c != '$')) {
      printf("\n  [CM] reading : [%2i] %s", i+1, all_seq_name[i]);
      fgets (junk, SIZE_MAX_NAME, incm);
      //  printf(" +%s+", junk);
      //  nnn;

      //      printf ("\t[%i.%i] %i ", i+1, j+1, nb_stems[10*i+j]);
      for (l=0; l<nb_stems[i]; l++) {
	fscanf (incm, "%i%i%i%i", &k, &x, &y, &length);
	//	printf ("%i %i %i %i\n", k, x, y, length);
	// printing base pairs
	for (m=0; m<length; m++) bp[x+m-1] = y-m;
      }

      // printing rnaml file
      m = strlen(all_seq_nucl[i]);
      fprintf (outrnaml," <xml>\n\n");
      fprintf (outrnaml," <rnaml>\n\n");
      fprintf (outrnaml," <rna name=\"%s\">\n\n", all_seq_name[i]);
      fprintf (outrnaml," <sequence><data>\n");
      //   begin sequence
      for (l=0; l<m; l++) {
	if (l%50==0) fprintf (outrnaml,"\n     ");
	fprintf (outrnaml,"%c", all_seq_nucl[i][l]);}
      //   end sequence
      fprintf (outrnaml,"\n\n </data></sequence>\n\n");
      fprintf (outrnaml," <structure>\n\n");
      //   begin structure
      for (l=0; l<m; l++) if (bp[l]) {
	fprintf (outrnaml," <base-pair _5p-base-num=\"%i\"", l);
	fprintf (outrnaml,"_3p-base-num=\"%i\"/>\n", bp[l]);}
      //   end structure
      fprintf (outrnaml,"\n\n </structure>\n\n");
      fprintf (outrnaml," </rnaml>\n\n");
      fprintf (outrnaml," </xml>\n\n");
    }
  }

  nnn; nnn;
  printf ("  forced stems printed in %s\n\n", name_out);
  fclose (outrnaml);
  fclose (incm);


  return 0;
}
