

/*------------------------------------------------------*/
/* accumulation.c                                       */
/*------------------------------------------------------*/

#include "accumulation.h"

/*------------------------------------------------------*/
void conglom_x (sequence_list_t all_seq, int x);
void final_structure_x (sequence_list_t all_seq, int x,
			int magnify, int TH1, int TH2);
stem_list_t cumul_freq (stem_list_t K, int seq);
/*------------------------------------------------------*/

void cat_stems (stem_list_t S, int x)
{extern stem_list_t *R; int i, n=stem_list_length (S);
 for (i=0; i<n; i++) add_stem_list (&(R[x]), S[i]);}

void conglomerate_stems (sequence_list_t all_seq)
{int i; extern int nb_seq;
 for(i=0; i<nb_seq; i++) conglom_x (all_seq, i);}

void conglom_x (sequence_list_t all_seq, int x) {
  extern stem_list_t *R; int n;
  // sort stems & cumulate frequencies
  n = stem_list_length (R[x]);
  sort_stems_by_begin (R[x]);
  R[x] = cumul_freq (R[x], x);
  sort_stems_by_freq (R[x]);}

void print_infos () {
  extern int nb_seq;
  extern sequence_list_t all_seq;
  FILE *out; int i, j;
  if (VERBOSE && CLEAN) line;
  out = fopen("output/infos.out","w");

  fprintf(out,"\n  AP_THD         %i",        AP_THD);
  fprintf(out,"\n  SUB            %i",           SUB);

  fprintf(out,"\n  SIZE_HAIRPIN   %i",  SIZE_HAIRPIN);

  fprintf(out,"\n  CORRECT_THD    %i",   CORRECT_THD);
  fprintf(out,"\n  INI_THD        %i",       INI_THD);
  fprintf(out,"\n  DIST_1         %i",        DIST_1);
  fprintf(out,"\n  THD_1          %i",         THD_1);
  fprintf(out,"\n  DIST_2         %i",        DIST_2);
  fprintf(out,"\n  THD_2          %i",         THD_2);

  fprintf(out,"\n  SINGLE_HP      %i",     SINGLE_HP);
  fprintf(out,"\n  SIZE_MAX_HP    %i",   SIZE_MAX_HP);
  fprintf(out,"\n  THD_HP         %i",        THD_HP);

  fprintf(out,"\n  FLT            %i",           FLT);

  fprintf(out,"\n  LARGE          %i",         LARGE);
  fprintf(out,"\n\n");

  for(i=0; i<nb_seq; i++) fprintf(out," [%2i] %s\n", i+1, all_seq[i].name);
    fprintf(out,"\n");
    for (j=10; j<=DIST_2; j+=10) 
      {fprintf(out," [%3i:%5i]", j, seuil(j, all_seq[0]));
      if (j%50==0) fprintf(out,"\n");}
    fprintf(out," [>%3i:%5i]", DIST_2, seuil(1000, all_seq[0]));
  fprintf(out,"\n\n _________________________________________________\n\n");
  fclose (out);
}

stem_list_t cumul_freq (stem_list_t K, int seq) {
  int i, j, n; stem_list_t L;
  n = stem_list_length (K);
  L = new_stem_list (n); j=0;
  if (n) { K[0]->seq = seq; L[0] = K[0];
    for (i=1; i<n; i++) { K[i]->seq = seq;
    if (same_stem (K[i-1], K[i])) (L[j]->freq)++;
    else {j++; L[j] = K[i];}}
    L[j+1] = NULL;}
  else L[0] = NULL;
  free (K); return L;}

/*------------------------------------------------------*/

void final_structure (sequence_list_t all_seq, int magnify, int TH1, int TH2)
{int i; extern int nb_seq;
 for(i=0; i<nb_seq; i++) final_structure_x (all_seq, i, magnify, TH1, TH2);}

void final_structure_x (sequence_list_t all_seq, int x,
			int magnify, int TH1, int TH2) {

  extern stem_list_t *R, *finalR;
  int i, j, n, N;
  char filename[20];
  FILE *out, *ct_out;
  string_t name;
  vector_t basep;
  int LIMIT_FREQ = 1;

  // finalR[x] will be the final list of stems for the seq x
  n = stem_list_length (R[x]);
  sort_stems_by_index (R[x]);
  finalR[x] = new_stem_list (n);


  // stepwise greedy incorporation by decreasing index

  for (i=0; i<n; i++)
    if (
	(R[x][i]->freq > LIMIT_FREQ) &&
	(R[x][i]->index >= TH1) &&
	(is_almost_compatible (R[x][i], finalR[x]))
	)
      {add_stem_list (&(finalR[x]), R[x][i]);}

  for (i=0; i<n; i++)
    if (
	(R[x][i]->freq <= LIMIT_FREQ) &&
	(R[x][i]->index >= TH2) &&
	(is_almost_compatible (R[x][i], finalR[x]))
	)
      {add_stem_list (&(finalR[x]), R[x][i]);}


/*    // nussinov incorporation by index */

/*    for (i=0; i<n; i++) */
/*      if (R[x][i]->folded) */
/*        {add_stem_list (&(finalR[x]), R[x][i]);} */



  // init file Z_* --> show list R[x]
  filename[0]='Z'; filename[1]='_';
  filename[2]='0'+((x+1)/100);
  filename[3]='0'+(((x+1)%100)/10);
  filename[4]='0'+((x+1)%10);
  filename[5]='.'; j=6;
  for (i=0; (i<10) && ((all_seq[x].name)[i]!='\0'); i++)
    if ((all_seq[x].name)[i]!=' ') {filename[j++]=(all_seq[x].name)[i];}
  filename[j++]='.'; filename[j++]='o'; filename[j++]='u';
  filename[j++]='t'; filename[j++]='\0';
  name =cat(new_string("output/"),new_string(filename));
  out = fopen(name,"w");

  // headings Z_*
  fprintf(out,"\n\n   name  : %s [%2i]\n", all_seq[x].name, x+1);
  fprintf(out,"   size  : %i nucl.\n", all_seq[x].length);
  fprintf(out,"   stems : %i\n\n", n);

  // contents  Z_*
  fprintf(out,"     flag        position      ");
  fprintf(out,"  freq   sig   rel  index     sub   sequence\n");
  fprintf(out,"____________________________________");
  fprintf(out,"_____________________________________\n");
  fprintf(out,"     |    |                      |    |     |     |     | | |   |\n");
  for(i=0; i<n; i++) {
    fprintf(out," %4i|%s |", i+1, int2string(R[x][i]->flag,1));
    fprintf(out,"%4i:%4i", R[x][i]->s[1]+1, R[x][i]->s[2]+1);
    fprintf(out," / %4i:%4i |", R[x][i]->s[3]+1, R[x][i]->s[4]+1);
    fprintf(out," %2i |", R[x][i]->freq);
    fprintf(out," %3i |", R[x][i]->sig);
    fprintf(out," %3i |", R[x][i]->rel);
    fprintf(out," %3i |", R[x][i]->index);
    // print star '*' if the stem belongs to finalR[x]
    if (belongs(R[x][i], finalR[x])) fprintf(out,"*|");
    else if (is_almost_compatible(R[x][i], finalR[x])) fprintf(out,"c|");
    else fprintf(out," |");
    // print star 'm' if the stem matches a true stem
    if (R[x][i]->true_match>=10) fprintf(out,"m|");
    else if (R[x][i]->true_match>0) fprintf(out,"s|");
    else fprintf(out," |");
    // highlight substems
    if (meet_substem(R[x][i], R[x]))
      fprintf (out, "sub|"); else fprintf (out, "   |");
    for(j=R[x][i]->s[1]; j<=R[x][i]->s[2]; j++)
      fprintf(out,"%c", all_seq[x].nucl[j]); fprintf(out," / ");
    for(j=R[x][i]->s[3]; j<=R[x][i]->s[4]; j++)
      fprintf(out,"%c", all_seq[x].nucl[j]); fprintf(out,"\n");}
  fprintf(out,"_______________________________________");
  fprintf(out,"_______________________________________\n");
  fclose (out);

  // init ct file --> finalR[x]
  n = all_seq[x].length;
  N = stem_list_length (finalR[x]);
  // A vector 'basep' is pre-computed. finalR[x] should be sorted by
  // decreasing significance so stems are incorporated in basep in that
  // order. If there is an overlap, the ancien stem has the priority
  // in the locus of the overlap; the remainder is filled by the new
  // stem, even if it stays only one base pair.
  basep  = new_vector (n);  basep[0] = 0;
  for (i=0; i<N; i++)
    for (j=finalR[x][i]->s[1]; j<=finalR[x][i]->s[2]; j++) {
      if (
	  (basep[j] == 0) &&
	  (basep[finalR[x][i]->s[4] + finalR[x][i]->s[1] - j] == 0)
	  ) {
	basep[j] = finalR[x][i]->s[4] + finalR[x][i]->s[1] - j + 1;
	basep[finalR[x][i]->s[4] + finalR[x][i]->s[1] - j] = j + 1;}
    }
  name = cat(cat(new_string("output/str."), intoa(x+1)), new_string(".xxx.ct"));
  ct_out = fopen(name,"w");

  // contents ct file
  fprintf (ct_out,"%5i %s\n", n, all_seq[x].name);
  for (i=0; i<n; i++) {
    fprintf (ct_out,"%5i ", i+1);
    //    fprintf (ct_out,"%c ", all_seq[x].nucl[i]);
    fprintf (ct_out,"%c ", toupper(all_seq[x].nucl[i]));
    fprintf (ct_out,"%5i ", i);
    fprintf (ct_out,"%5i ", (i+2)%(n+1));
    fprintf (ct_out,"%5i ", basep[i]);
    fprintf (ct_out,"%5i\n", i+1);}
  fclose (ct_out);

  // magnify energy
  if (magnify) for (i=0; i<N; i++) finalR[x][i]->energy *= finalR[x][i]->index;

}
