
/* -------------------------------------------*/
/* clique.c                                   */
/*--------------------------------------------*/

#include <unistd.h>
#include "clique.h"

extern sequence_list_t all_seq;
extern stem_list_t *R;
extern int nb_seq;
g_node_list_t g_node;
compound_list_t compound_list;

/*--------------------------------------------*/
void gather_cofolded_seq (int i);
void display_cofolded_seq ();
void find_equiv_stems (stem_list_t S, int i_seq, int si);
void create_g_node (stem_t P, int t);
void connect_g_node (g_node_t G);
void create_compound (g_node_t G, int t);
void add_id_edges (compound_t C);
void delete_arity_one (compound_t C);
void clear_conficting_nodes (compound_t C);
/*--------------------------------------------*/

void build_graphs () {

  int i, j, k, n, t;
  int nb_gnodes, nb_compounds;
  vector_t ns;
  g_node_t G;


  // gather cofolded sequences
  for (i=0; i<nb_seq; i++) gather_cofolded_seq (i);
  display_cofolded_seq ();


  // allocate memory
  ns = new_vector (nb_seq);
  g_node = new_g_node_list (RATIO_NB_MAX_G_NODES*nb_seq);
  compound_list = new_compound_list (RATIO_NB_MAX_COMPOUNDS*nb_seq);


  // find equivalent stems
  // they are stored in a field of stem_t
  // + refresh field 'tag' & 'g_node' of each stem
  for (i=0; i<nb_seq; i++) {
    ns[i] = stem_list_length (R[i]);
    for (j=0; j<ns[i]; j++) {
      R[i][j]->tag=0;
      R[i][j]->g_node=-1;
      if (R[i][j]->equiv) free (R[i][j]->equiv);
      find_equiv_stems (R[i], i, j);}}


  // create g_nodes
  t=0; if (VERBOSE) printf ("\n");
  for (i=0; i<nb_seq; i++)
    for (j=0; j<ns[i]; j++) {
      create_g_node (R[i][j], t);
      g_node[t] -> seq = i;
      if (stem_list_length (g_node[t]->in_stem_list) != 0) t++;
      if (t>=RATIO_NB_MAX_G_NODES*nb_seq)
	{printf("  too many nodes\n\n");
	fprintf(stderr, "  (RATIO_NB_MAX_G_NODES = %i\n", RATIO_NB_MAX_G_NODES);
	exit(0);}}
  g_node[t] = NULL;
  nb_gnodes = g_node_list_length (g_node);


  // print classes
  if (DEBUG) for (i=0; i<nb_gnodes; i++) print_class_content (g_node, i);


  // connect g_nodes
  for (i=0; i<nb_gnodes; i++) {
    G = g_node[i];
    if (!G->visited) connect_g_node (G);}
  for (i=0; i<nb_gnodes; i++) {
    g_node[i]->visited=0;
    sort_out_g_node_list (g_node[i]);}


  // create connex compounds (= transitive closures)
  t=0; for (i=0; i<nb_gnodes; i++) {
    G = g_node[i];
    if (!G->visited) {
      if (t >= RATIO_NB_MAX_COMPOUNDS*nb_seq-1)
	{fprintf(stderr, "  out of range in build_graph\n");
	fprintf(stderr, "  (RATIO_NB_MAX_COMPOUNDS = %i\n", RATIO_NB_MAX_COMPOUNDS);
	exit(0);}
      add_g_node_to_compound (compound_list[t], G);
      create_compound (G, t);
      if (compound_list[t]->size) t++;
    }
  }

  nb_compounds = t;
  compound_list [nb_compounds] = NULL;


  // no graph is found (all the graphs are too large)
  if (nb_compounds==0)
    {if (VERBOSE) printf("  no structure was found by CARNAC\n\n"); exit(0);}


  // complete equivalent edges when the stems
  // have not been cofolded because they are identical
  for (j=0; j<nb_compounds; j++) add_id_edges (compound_list[j]);


  // recursively delete arity 1 (in each compound)
  //  for (j=0; j<nb_compounds; j++) delete_arity_one (compound_list[j]);


  // sort nodes (in each compound)
  //  for (j=0; j<nb_compounds; j++) sort_list_compound (compound_list[j]);


  // print graph informations before cleaning
  //  for (j=0; j<nb_compounds; j++) print_graph_infos (compound_list[j]);
  //  close_graph_infos ();


  // clear conflicting nodes (in each compound)
  for (j=0; j<nb_compounds; j++) clear_conficting_nodes (compound_list[j]);


  // sort all links (in each node)
  for (i=0; i<nb_gnodes; i++) sort_out_g_node_list (g_node[i]);


  // compute graph caracteristics
  for (j=0; j<nb_compounds; j++) compute_stats (compound_list[j]);


  // compute 'sig' & 'rel' for each stem
  for (i=0; i<nb_gnodes; i++) {
    G = g_node[i];
    n = stem_list_length (G->in_stem_list);
    for (k=0; k<n; k++) {
      if (G -> compound >= 0) {
	G->in_stem_list[k] -> index =
	  compound_list [G -> compound] -> index;
	G->in_stem_list[k] -> sig =
	  compound_list [G -> compound] -> sig;
	G->in_stem_list[k] -> rel =
	  compound_list [G -> compound] -> rel;}
      else {
	// single stem compound ---> -1
	// oor in compound ---> -2
	G -> in_stem_list[k] -> sig = G -> compound;
	G -> in_stem_list[k] -> rel = G -> compound;
	G -> in_stem_list[k] -> index = G -> compound;
      }}}


  // sort the compound list according to their significance
  // WARN: !! field  g_node -> compound not refreshed !!
  //  sort_compounds_list_by_sig (compound_list, nb_compounds);


  // sort the compound list according to their reliability
  // WARN: !! field  g_node -> compound not refreshed !!
  //  sort_compounds_list_by_rel (compound_list, nb_compounds);


  // sort the compound list according to their index
  // WARN: !! field  g_node -> compound not refreshed !!
  sort_compounds_list_by_index (compound_list, nb_compounds);


  // print  &  display
  if (VERBOSE) {printf("  [%i classes of stems in %i", nb_gnodes, nb_compounds);
  printf(" connex compounds]\n\n");}
  if (DEBUG) for (i=0; i<nb_gnodes; i++) print_class_connect (g_node, i);
  for (j=0; j<nb_compounds; j++) {
    sort_list_compound (compound_list[j]);
    if (DEBUG) print_connex_compounds (compound_list, j);
    if (GRAPH) generate_dot_file (compound_list, j);
    print_graph_infos (compound_list[j]);
  }
  close_graph_infos ();


  // refresh field 'married' of each stem
  for (i=0; i<nb_seq; i++)
    for (j=0; j<ns[i]; j++)
        for (k=0; k<nb_seq; k++)
	  R[i][j]->married[k] = NULL;


  // free memory
  free_g_node_list (g_node, g_node_list_length(g_node));
  free_compound_list (compound_list, nb_compounds);
  free (ns);
}


void gather_cofolded_seq (int i) {

  int flag, j, k;

  if (all_seq[i].N == -1) {
    // seq #i hasn't been visited yet
    all_seq[i].N = -2;
    // now it has been
    flag = 1;
    while (flag) {
      flag = 0;
      for (j=0; j<nb_seq; j++)
	if (
	    (all_seq[i].cofold[j]) &&
	    (all_seq[j].N == -1)
	    ) {
	  all_seq[j].N = -2;
	  for (k=0; k<nb_seq; k++) {
	    if (
		(k != i) &&
		(!all_seq[i].cofold[k]) &&
		(all_seq[j].cofold[k])
		) {
	      all_seq[i].cofold[k] = 2;
	      flag = 1;
	    }
	  }
	}
    }
    all_seq[i].N = 0;
    for (j=0; j<nb_seq; j++)
      if (all_seq[i].cofold[j]) all_seq[i].N++;
  }

  else if (all_seq[i].N == -2) {
    // seq #i has already been visited
    for (j=0; ((!all_seq[j].cofold[i]) || (all_seq[j].N < 0)) && (j<nb_seq); j++);
    for (k=0; k<nb_seq; k++)
      if (
	  (!all_seq[i].cofold[k]) &&
	  (all_seq[j].cofold[k])
	  )  all_seq[i].cofold[k] = 2;
    all_seq[i].N = all_seq[j].N;
  }
}

void display_cofolded_seq () {
  FILE *out; int i, j, cpt;
  out = fopen ("output/cofoldseq.out", "w");
  fprintf (out, "\n    ");
  for (j=0; j<nb_seq; j++) fprintf (out, "%2i ", j+1); fprintf (out, "\n\n");
  for (i=0; i<nb_seq; i++) {
    fprintf (out, "%2i: ", i+1);
    cpt = 0;
    for (j=0; j<nb_seq; j++) {
      if (i==j) fprintf (out, "  ");
      else if (all_seq[i].cofold[j] == 1) {cpt++; fprintf (out, "  ");} // cofold
      else if (all_seq[i].cofold[j] == 2) fprintf (out, " . ");  // trans. closure
      else fprintf (out, "   ");}
    fprintf (out, " (%2i ->", cpt);
    fprintf (out, " %2i) ", all_seq[i].N);
    fprintf (out, " %s\n", all_seq[i].name);}
  fprintf (out, "\n");
  fclose (out);
}


void find_equiv_stems (stem_list_t S, int i_seq, int si)
{int i, t, n; stem_t P; n=stem_list_length (S); P=S[si];
 P -> equiv = new_stem_list (SIZE_EQUIV);
 t = 0; for (i=0; i<n; i++) {
   if (
       (equiv_stem (P, S[i])) &&
       //       (S[i] -> freq > 1) && (P -> freq > 1) &&
       (i!=si) &&
       (t<SIZE_EQUIV)
       ) {
     //     print_stem (P); print_stem (S[i]); nnn;
     P -> equiv[t] = S[i]; t++;}} P -> equiv[t] = NULL;}


void create_g_node (stem_t P, int t) { 
  int i, n;
  if (P->tag == 0) {
    P->tag = 1; P->g_node = t;
    add_stem_list (&(g_node[t]->in_stem_list), P);
    //    printf ("[%s]: ", int2string (t,0)); print_stem (P);
    n = stem_list_length(P->equiv);
    for (i=0; i<n; i++)
      create_g_node (P->equiv[i], t);}}


void connect_g_node (g_node_t G) {
  int i, j, n;
  stem_t P, Pe;
  g_node_t out;
  if (!G->visited) {
    G -> visited = 1;
    n = stem_list_length (G->in_stem_list);
    for (i=0; i<n; i++) {
      P = G->in_stem_list[i];
      for (j=0; j<nb_seq; j++) {
	Pe = P->married[j];
	if (Pe) {
	  out = g_node[Pe->g_node];
	  add_to_out_g_node_list (G, out);
	  connect_g_node (out);}
      }}}}

void clean_compound (g_node_t G) {
  int i, n; g_node_t Gout;
  G -> visited = 2;
  G -> compound = -2;
  n = G->arity_out;
  for (i=0; i<n; i++) {
    Gout = G->out_g_node_list[i];
    if (Gout->visited != 2)
      clean_compound (Gout);}}


void create_compound (g_node_t G, int t) {
  int i, n, too_large=0; g_node_t Gout;
  if (!G->visited) {
    G -> visited = 1;
    G -> compound = t;
    n = G->arity_out;
    for (i=0; i<n; i++) {
      Gout = G->out_g_node_list[i];
      if (!Gout->visited) {
	too_large += add_g_node_to_compound (compound_list[t], Gout);
	if (too_large) clean_compound (G);
	else create_compound (Gout, t);
      }
    }
  }
}


void add_id_edges (compound_t C) {
  int i, ii, j, jj, n, nn;
  int witness;
  stem_t P, PP;
  g_node_t G, GG;
  for (j=0; j<C->size; j++) {
    G = C->list[j];
    n = stem_list_length (G->in_stem_list);
    for (jj=j+1; jj<C->size; jj++) {
      GG = C->list[jj];
      if ((G->seq != GG->seq) &&
	  (!belongs2g_node_list(G,GG->out_g_node_list))) {
	witness=0;
	for (i=0; i<n; i++) {
	  P = G->in_stem_list[i];
	  nn = stem_list_length (GG->in_stem_list);
	  for (ii=0; ii<nn; ii++) {
	    PP = GG->in_stem_list[ii];
	    if (
		(!witness) &&
		(P->married[PP->seq] != PP) &&
		//(identity (P, PP))
		(uncovary (P, PP))
		) {
	      // print_stem (P); print_stem (PP); nnn;
	      witness = 1;
	      add_to_id_g_node_list (G, GG);
	      add_to_id_g_node_list (GG, G);}}}}}}}


void delete_arity_one (compound_t C) {
  int witness=1; int i, j;
  g_node_t G, Gid, Gout;
  g_node_list_t S;
  while (witness) {
    witness = 0;
    for (i=0; i<C->size; i++) {
      G = C->list[i];
      if (G->arity_out==1) {
	witness = 1;
	Gout = G->out_g_node_list[0];
	S = Gout->out_g_node_list;
	delete_g_node (S, rank(S,G), G);
	Gout -> arity_out --;
	for (j=0; j<G->arity_id; j++) {
	  Gid = G->id_g_node_list[j];
	  S = Gid->id_g_node_list;
	  delete_g_node (S, rank(S,G), G);
	  Gid -> arity_id --;}
	G -> arity_id = 0;
/*  	printf ("[%s] deleted ", int2string (G->rank,0)); */
/*  	printf ("[%s]->arity_out top down to %i\n", */
/*  		int2string (Gout->rank,0), Gout->arity_out); */
	delete_g_node (C->list, i, G);
	C -> size --;}}}}


void clear_conficting_nodes (compound_t C) {
  int witness; int i, j, min_arity;
  g_node_t G1, G2, Gid, Gout; g_node_list_t S;
  int ARITY = nb_seq/4;
  for (min_arity=1; min_arity<=ARITY; min_arity++) {
    witness = 1;
    while (witness) {
      sort_list_compound (C);
      //      printf ("--> [%s] (size=%i)\n\n", int2string(C->list[0]->rank,0),C->size);
      witness = 0;
      for (i=0; i<C->size-1; i++) {
	G1 = C->list[i];
	G2 = C->list[i+1];
	if (G1->seq == G2->seq) {
	  // delete G1
	  if ((G1->arity_out <= min_arity) &&
	      (G2->arity_out >= min_arity+1)) {
	    witness = 1;
	    //	    printf ("\n[%2i:%s] deleted\n", G1->seq+1, int2string (G1->rank,0));
	    for (j=0; j<G1->arity_id; j++) {
	      // delete all incoming 'id' edges
	      Gid = G1->id_g_node_list[j];
	      S = Gid->id_g_node_list;
	      delete_g_node (S, rank(S,G1), G1);
	      Gid -> arity_id --;
	      //	      printf ("[%2i:%s]->arity_id top down to %i\n", Gid->seq+1,
	      //		      int2string (Gid->rank,0), Gid->arity_id);
	    }
	    for (j=0; j<G1->arity_out; j++) {
	      // delete all incoming 'out' edges
	      Gout = G1->out_g_node_list[j];
	      S = Gout->out_g_node_list;
	      delete_g_node (S, rank(S,G1), G1);
	      Gout -> arity_out --;
	      //	      	      printf ("[%2i:%s]->arity_out top down to %i\n", Gout->seq+1,
	      //	      		      int2string (Gout->rank,0), Gout->arity_out);
	    }
	    G1 -> compound = -1;
	    //	    G1 -> arity_id = 0;
	    //	    G1 -> arity_out = 0; 
	    delete_g_node (C->list, i, G1);
	    C -> size --;}
	  // delete G2
	  else if ((G2->arity_out <= min_arity) &&
		   (G1->arity_out >= min_arity+1)) {
	    witness = 1;
	    //	    	    printf ("\n[%2i:%s] deleted\n", G2->seq+1, int2string (G2->rank,0));
	    for (j=0; j<G2->arity_id; j++) {
	      // delete all incoming 'id' edges
	      Gid = G2->id_g_node_list[j];
	      S = Gid->id_g_node_list;
	      delete_g_node (S, rank(S,G2), G2);
	      Gid -> arity_id --;
	      //	      	      printf ("[%2i:%s]->arity_id top down to %i\n", Gid->seq+1,
	      //	      		      int2string (Gid->rank,0), Gid->arity_id);
	    }
	    for (j=0; j<G2->arity_out; j++) {
	      // delete all incoming 'out' edges
	      Gout = G2->out_g_node_list[j];
	      S = Gout->out_g_node_list;
	      delete_g_node (S, rank(S,G2), G2);
	      Gout -> arity_out --;
	      //	      	      printf ("[%2i:%s]->arity_out top down to %i\n", Gout->seq+1,
	      //	      		      int2string (Gout->rank,0), Gout->arity_out);
	    }
	    G2 -> compound = -1;
	    //	    G2 -> arity_id = 0;
	    //	    G2 -> arity_out = 0;
	  delete_g_node (C->list, i+1, G2);
	  C -> size --;
	  }
	}
      }
    }
  }
}

void get_remaining_stems () {
  extern stem_list_t *finalR, *remain;
  extern sequence_list_t all_seq;
  extern stem_list_t *all_stems;
  extern int nb_seq;
  int nb_cofolds;
  stem_list_t K;
  int x, i, n, nr, nf;

  for (x=0; x<nb_seq; x++) {
    K = all_stems[x];
    n = stem_list_length (K);
    nb_cofolds = -1;
    for(i=0; i<nb_seq; i++)
      nb_cofolds += (all_seq[x].cofold[i]==1);
    for(i=0; i<n; i++)
      if (is_almost_compatible (K[i],finalR[x]))
	add_stem_list (&(remain[x]), K[i]);
    nr = stem_list_length (remain[x]);
    nf = stem_list_length (finalR[x]);
    if (VERBOSE)
      printf ("[%2i] [%2i cofoldings] [%4i stems -> %4i selected + %4i remaining (%2i%c)]\n",
	      x+1, nb_cofolds, n, nf, nr-nf, 100*(nr-nf)/n, '%');
    print_remaining_stems (all_seq, x, remain[x]);}

}
