

/* -------------------------------------------*/
/* find_blocks.c                              */
/*--------------------------------------------*/

#include "find_blocks.h"

extern block_list_t block;

/*---------------------------------------------------*/
void check_consistency (block_list_t block);
void erase_block (int l);
void unoverlap_blocks (int l);
void sort_blocks (block_list_t block);
/*--------------------------------------------------*/

block_list_t find_blocks (sequence_list_t seq) {

  extern matrix2_t bino;
  extern int brk_lp;
  int d, i, j, k, l, p, t, n, m;
  int length_alloc;
  int common_length, score, score_back, top_score, seuil;
  char a, b;
  matrix2_t aux;
  int begin0, begin1;

  seuil = AP_THD;

  n = seq[0].length;
  m = seq[1].length;

  length_alloc = (n+m);
  block = alloc_block_list (NB_MAX_BLOCKS);
  aux = alloc_matrix2 (length_alloc, 2, -1);

  // compute and store binomial coefs
  bino = alloc_matrix2 (32, 32, 0);
  bino[0][0] = 1; for (i=1; i<32; i++) {
    bino[i][0] = 1; for (j=1; j<i; j++)
      bino[i][j] = bino[i-1][j] + bino[i-1][j-1];}

  // blocs in situ : on repre les zones de similarit
  // dont la fin est stocke dans une matrice auxiliaire
  //  0 :   i i i i i i i i ... i i  -1  -1  -1
  //  1 :   j j j j j j j j ... j j  -1  -1  -1
  // les squences sont parcourues de droite  gauche.
  // la squence 0 est prise comme rfrence, et on fait
  // 'coulisser' la squence 1.

  k = -1;
  // 'd' est le dcalage de la sq. 1 par rapport  la sq. 0
  for (d=-m; d<=m; d++) {
    score = 0;
    top_score = 0;
    common_length = MIN (m+d, MIN (n, MIN (n-d, m)));
    for (i=common_length; i>=0; i--) {

      // top_score est la valeur du dernier maximum local.
      // chaque fois que le score descend sous les 2/3
      // du seuil, top_score est rinitialis.
      if (3*score < 2*seuil) top_score = 0;

      // si top_score est positif, chaque fois que le score
      // a dvalu de 1/3 de top_score, le score est rinitialis.
      if ((top_score) && ((top_score-score) > seuil)) score = 0;

      a = seq[0].nucl[i + MAX(d,0)];
      b = seq[1].nucl[i - MIN(d,0)];

      //      if (d==-1) print_(score);

      // on arrive tout  gauche :
      // on vrifie le score.
      if ((i == 0) &&
	  (
	   ((top_score) && (score > top_score)) ||
	   (score >= seuil))
	  )

	{// si top_score avait t rinitialis,
	  // on stocke le maximum local.
	  // sinon, on remplace celui qu'on avait
	  // stock avant.
	  if (top_score == 0) k++;

/*  	  printf ("* %i (%i) %i:%i\n", top_score, score, */
/*  		  i + MAX(d,0),i - MIN(d,0)); */

	  if (SUBST(a,b) > 0) {
	    aux [k][-1] = score + 1;
	    aux [k][0]  = i + MAX(d,0);
	    aux [k][1]  = i - MIN(d,0);}
	  else {
	    aux [k][-1] = score;
	    aux [k][0]  = i + MAX(d,0) + 1;
	    aux [k][1]  = i - MIN(d,0) + 1;}}

      // on vrifie le score
      else if ((SUBST(a,b) < 0) &&
	       (score >= seuil) &&
	       (score > top_score))

	{if (top_score == 0) k++;

/*  	printf ("  %i (%i) %i:%i\n", top_score, score, */
/*  		i + MAX(d,0)+1,i - MIN(d,0)+1); */

	top_score   = score;
	aux [k][-1] = score;
	aux [k][0]  = i + MAX(d,0) + 1;
	aux [k][1]  = i - MIN(d,0) + 1;}

      score = MAX(0, score + SUBST (a,b));
    }

    if (k==length_alloc-1)
      {printf("  change : length_alloc\n"); exit(1);}
  }

/*    for (k=0; k<100; k++) if (aux [k][-1] >= 10) */
/*      printf (" [%i] %i.%i.%i\n", k, aux [k][-1], aux [k][0], aux [k][1]); */

  // on retrace les blocs

  l = 0;
  for (t=0; aux[t][0]!=-1; t++);
  for (k=t-1; k>=0; k--) {
    if (aux[k][0]!=-5) {
      i = aux[k][0]; j = aux[k][1];
      // nouveau bloc
      if (l==NB_MAX_BLOCKS-2)
	{printf("\n -- too many blocks --\n"); exit(1);}
      block[l] = new_block (i, j, 0, 0);
      score_back = aux[k][-1];
      common_length = MIN (n-1 - i, m-1 - j);
      for (d=0; (score_back>0) && (d<common_length); d++) {
	a = seq[0].nucl[i+d];
	b = seq[1].nucl[j+d];
	score_back -= SUBST (a,b);
	if (SUBST(a,b)!=1) block[l]->mism++;
	// vrifie les duplicats
	for (p=k-1; p>=0; p--)
	  if ((aux[p][0] == i+d) && (aux[p][1] == j+d))
	    {aux[p][0]=-5; aux[p][1]=-5;}
      }
      block[l]->length = d;
      if (d == common_length) (block[l]->length)++;
      // en fonction de la probabilit d'apparition du bloc,
      // on dcide si on le garde ou non
      if (l==0) {
	begin0 = block[l]->begin[0];
	begin1 = block[l]->begin[1];
      }
      else {
	begin0 = block[l]->begin[0] - block[l-1]->begin[0];
	begin1 = block[l]->begin[1] - block[l-1]->begin[1];
      }
      if (rel_weight (begin0, begin1, block[l]->length, block[l]->mism)
	  >= SEUIL_PROBA_BLOCKS) l++;
    }
  }
  block[l]=NULL;

  sort_blocks (block);
  //  print_all_blocks (seq, block);
  free_matrix2 (aux, length_alloc);
  if (DEBUG) printf("[%2i]", block_list_length (block)); fflush (stdout);
  check_consistency (block);
  if (VERBOSE) printf("[%2i anchors] ", block_list_length (block)); fflush (stdout);
  if (block_list_length (block) == 0) brk_lp++;

  return block;
}

void erase_block (int l)
{int i; for (i=l; block[i]!=NULL; i++) block[i] = block[i+1];}

void unoverlap_blocks (int l) {

  extern int i, j;
  extern sequence_list_t seq;
  int overlap, length0, length1;
  FILE *out=NULL;

  length0 = seq[0].length;
  length1 = seq[1].length;

  if (DEBUG) out=fopen("output/check_blocks.out","a");

  // si le 2me bloc est inclus dans le premier, on le supprime

  if ((block[l+1]->begin[0] + block[l+1]->length <=
      block[l]->begin[0] + block[l]->length) ||
      (block[l+1]->begin[1] + block[l+1]->length <=
       block[l]->begin[1] + block[l]->length)) {

    if (DEBUG) {
      fprintf(out, "\n\n[%2i:%2i] nested blocks\n [%3i..%3i] .. [%3i..%3i]\n",
	      i+1, j+1,
	      block[l]->begin[0]+1,
	      block[l]->begin[0]+block[l]->length,
	      block[l+1]->begin[0]+1,
	      block[l+1]->begin[0]+block[l+1]->length);
      fprintf(out," [%3i..%3i] .. [%3i..%3i]\n\n",
	      block[l]->begin[1]+1,
	      block[l]->begin[1]+block[l]->length,
	      block[l+1]->begin[1]+1,
	      block[l+1]->begin[1]+block[l+1]->length);
      fprintf(out,"\n[%2i:%2i] block [%i.%i.%i.%i] deleted\n\n",
	      i+1, j+1,
	      block[l+1]->begin[0]+1,
	      block[l+1]->begin[0]+block[l+1]->length,
	      block[l+1]->begin[1]+1,
	      block[l+1]->begin[1]+block[l+1]->length);
    }

    erase_block (l+1);

  }
  else {

  // sinon on supprime la partie commune aux deux blocs

    overlap = MAX(block[l]->begin[0] + block[l]->length -
		  block[l+1]->begin[0]+1,
		  block[l]->begin[1] + block[l]->length -
		  block[l+1]->begin[1]+1);

    if (DEBUG) {
      fprintf(out,"\n\n[%2i:%2i] overlapping blocks: overlap %i\n",
	      i+1, j+1,overlap+1);
      fprintf(out," [%3i..%3i] // [%3i..%3i]\n",
	      block[l]->begin[0]+1,
	      block[l]->begin[0]+block[l]->length,
	      block[l+1]->begin[0]+1,
	      block[l+1]->begin[0]+block[l+1]->length);
      fprintf(out," [%3i..%3i] // [%3i..%3i]\n",
	      block[l]->begin[1]+1,
	      block[l]->begin[1]+block[l]->length,
	      block[l+1]->begin[1]+1,
	      block[l+1]->begin[1]+block[l+1]->length);
    }

    block[l]->length -= overlap;
    block[l]->length = MIN (block[l]->length,
			      length0 - block[l]->begin[0] + 1);
    block[l]->length = MIN (block[l]->length,
			      length1 - block[l]->begin[1] + 1);
    
    block[l+1]->begin[0] += overlap+1;
    block[l+1]->begin[1] += overlap+1;
    block[l+1]->length -= overlap+1;

    if (DEBUG) {
      fprintf(out," --> [%3i..%3i] .. [%3i..%3i]\n",
	      block[l]->begin[0]+1,
	      block[l]->begin[0]+block[l]->length,
	      block[l+1]->begin[0]+1,
	      block[l+1]->begin[0]+block[l+1]->length);
      fprintf(out," --> [%3i..%3i] .. [%3i..%3i]\n\n",
	      block[l]->begin[1]+1,
	      block[l]->begin[1]+block[l]->length,
	      block[l+1]->begin[1]+1,
	      block[l+1]->begin[1]+block[l+1]->length);
    }
  }
  if (DEBUG) fclose(out);
}

void check_consistency (block_list_t block) {

  extern int i, j;
  int k, l, n, esp[2];
  int witness;
  FILE *out=NULL;

  if (DEBUG) out=fopen("output/check_blocks.out", "a");

  // vrifie que l'ancrage sur les blocs est cohrent

  witness = 1;

  while (witness) {
    n = block_list_length (block);
    witness = 0;
    for (l=0; l<n-1; l++) {
      n = block_list_length (block);
      // blocs croiss
      if ((block[l] != NULL) && (block[l+1] != NULL) &&
	  ((block[l]->begin[1] > block[l+1]->begin[1]) ||
	   ((block[l]->begin[0] == block[l+1]->begin[0]) &&
	    (block[l]->begin[1]+block[l]->length <= block[l+1]->begin[1]))))

	{
	  if (DEBUG) {
	    fprintf(out,"\n\n[%2i:%2i] crossed blocks\n [%3i..%3i] xx [%3i..%3i]\n",
		    i+1, j+1,
		    block[l]->begin[0]+1,
		    block[l]->begin[0]+block[l]->length,
		    block[l+1]->begin[0]+1,
		    block[l+1]->begin[0]+block[l+1]->length);
	    fprintf(out," [%3i..%3i] xx [%3i..%3i]\n",
		    block[l]->begin[1]+1,
		    block[l]->begin[1]+block[l]->length,
		    block[l+1]->begin[1]+1,
		    block[l+1]->begin[1]+block[l+1]->length);
	  }

	  for (k=l; k<=l+1; k++) {

	    if (k) esp[k-l] = rel_weight (block[k]->begin[0]-block[k-1]->begin[0],
					  block[k]->begin[1]-block[k-1]->begin[1],
					  block[k]->length, block[k]->mism);
	    else esp[k-l] = rel_weight (block[k]->begin[0], block[k]->begin[1],
					block[k]->length,block[k]->mism);

	    if (DEBUG) {
	      fprintf (out, "[length: %i ", block[k]->length);
	      fprintf (out, "mism: %i ", block[k]->mism);
	      fprintf (out, "dist: %i ", block[k]->begin[0]-block[k]->begin[1]);
	      fprintf (out, "pr: %i ",
		       rel_weight (0, 0, block[k]->length, block[k]->mism));
	      fprintf (out, "esp: %i]\n", esp[k-l]);
	      fprintf (out, "\n");}

	  }

	  if (esp[0] < esp[1]) erase_block (l);
	  else if ((esp[0] == esp[1]) &&
		   (block[l]->length < block[l+1]->length)) erase_block (l);
	  else erase_block (l+1);
	  witness++;
	}

      // blocs chevauchants
      else if ((block[l] != NULL) && (block[l+1] != NULL) &&
	       ((block[l]->begin[0] + block[l]->length >= block[l+1]->begin[0]) ||
		(block[l]->begin[1] + block[l]->length >= block[l+1]->begin[1])))
	{unoverlap_blocks (l); witness++;}

      // on vrifie qu'il reste un bout du bloc de taille
      // suprieure   5  sinon on le supprime

      if (block[l]->length <= 5)
	{
	  if (DEBUG) {
	    fprintf(out,"\n[%2i:%2i] block length < 5 [%i.%i.%i.%i] deleted\n\n",
		    i+1, j+1,
		    block[l]->begin[0]+1,
		    block[l]->begin[0]+block[l]->length,
		    block[l]->begin[1]+1,
		    block[l]->begin[1]+block[l]->length);
	  }
	  erase_block (l);
	}
    }
  }
  if (DEBUG) fclose(out);
}

void sort_blocks (block_list_t block) {
  block_t tmp; int l, n, witness;
  // trie la liste des blocs par rapport  begin[0]
  n = block_list_length (block); witness = 1;
  while (witness) {
    witness = 0; for (l=0; l<n-1; l++) {
      if (block[l]->begin[0] > block[l+1]->begin[0])
	{witness++; tmp = block[l];
	block[l] = block[l+1]; block[l+1] = tmp;}}}
}
