
/*----------------------------------------------*/
/* get_sequences.c                              */
/*----------------------------------------------*/

#include "get_sequences.h"

vector_t true_bp;
int ave_true_bp;
vector_t correct_stems_in;
int ave_correct_stems_in;
stem_list_t *true_stems;
int discarded;
vector_t discd;


void erase_out (){
  //  FILE *connect, *class_connect, *class_content;
  FILE *graph;

  fclose(fopen("output/reference.out","w"));
  fclose(fopen("output/true.out","w"));
  fclose(fopen("output/infos.out","w"));
  fclose(fopen("output/compare.out","w"));
  if (OUTPUT) fclose(fopen("output/folding.out","w"));
  if (OUTPUT) fclose(fopen("output/alignment.out","w"));

  graph = fopen("output/graph.out","w");
  fprintf (graph, "\n\n ** informations about the connex compounds **\n\n");
  fprintf (graph, "N: number of nodes\n");
  fprintf (graph, "Ns: number of sequences present in the graph\n");
  fprintf (graph, "sq: number of sequences\n");
  fprintf (graph, "n: number of stems\n");
  fprintf (graph, "\n");
  fprintf (graph, "co: number of edges 'cofolding'\n");
  fprintf (graph, "id: number of edges 'identity'\n");
  fprintf (graph, "tt: total number of edges\n");
  fprintf (graph, "me: max possible number of edges\n");
  fprintf (graph, "\n");
  //  fprintf (graph, "s: significance = (2*Ns-N)/sq * (me-id)/me\n");
  //  fprintf (graph, "r: reliability = co/(me-id)\n");
  fprintf (graph, "i: index = (2*Ns-N)/sq * co/(me-id)\n");
  fprintf (graph, "\n");
  fprintf (graph, "**: the compound is correct (T=N)\n");
  fprintf (graph, " *: the compound is almost correct (T>=N/2)\n");
  fprintf (graph, " -: there are correct stems\n");
  fprintf (graph, "    but the compound is too small (t>0) & (N<3)\n");
  fprintf (graph, " o: the stems are false while the compound\n");
  fprintf (graph, "    is rather big (t<N/2) & (N>3)\n");
  fprintf (graph, "\n");
  fprintf (graph, "    |    N    |    n    | sq | co | id | tt |  me ||  i |\n");
  fprintf (graph, "----------------------------------------------------------------------\n");
  fclose (graph);

/*    class_content = fopen("output/class_content.out","w"); */
/*    fprintf (class_content, "\n\n ** class content: list of stems ** \n"); */
/*    fclose (class_content); */

/*    class_connect = fopen("output/class_connect.out","w"); */
/*    fprintf (class_connect, "\n\n ** for each class, the list "); */
/*    fprintf (class_connect, "of directly connected classes ** \n"); */
/*    fclose (class_connect); */

/*    connect = fopen("output/connect.out","w"); */
/*    fprintf (connect, "\n\n ** connex compounds ** \n"); */
/*    fclose (connect); */

}

sequence_list_t get_sequences (string_t infile_seq) {

  extern matrix2_t similarity;
  extern sequence_list_t seq;
  extern string_t refname;
  extern int nb_seq;
  extern stem_list_t *R, *finalR, *remain;
  string_t base = "AUCGaucg";
  sequence_list_t all_seq;
  int i, j, k, n, skip, aa, uu, cc, gg, au;
  char c;
  FILE *f;

  // reset output files
  erase_out ();

  // control: filename '.fasta'
  if (infile_seq[strlen(infile_seq)-1] == '.')
    infile_seq = cat (infile_seq, new_string("fasta"));
  if (strcmp((infile_seq)+strlen(infile_seq)-6,".fasta"))
    infile_seq = cat (infile_seq, new_string(".fasta"));
  // control: existing file ?
  if ((f=fopen(infile_seq,"r")) == NULL)
    {fprintf(stderr, "  unknown file %s\n", infile_seq);
    exit(0);}
  // store refname
  refname = new_string (infile_seq);
  refname[strlen(infile_seq)-6] = '\0';
  if (VERBOSE) printf ("\n  reading sequence file: %s.fasta\n", refname);

  // read sequences

  discarded = 0;
  nb_seq = 0; while ((c=fgetc(f))!=EOF)
    {if (c=='>') nb_seq++; if (c=='$') nb_seq--;}
  discd = new_vector (nb_seq); discd[0] = 0;
  all_seq = (sequence_list_t) malloc (nb_seq * sizeof(sequence_t));
  seq = (sequence_list_t) malloc( 2 * sizeof(sequence_t));
  f=fopen(infile_seq,"r");
  while((c != '>') && (c != '$') && (c != EOF)) c=fgetc(f);
  c=fgetc(f); for (j=0; j<nb_seq; j++) {
    similarity = alloc_matrix2 (nb_seq, nb_seq, -1);
    all_seq[j].nucl = (string_t) calloc(SIZE_MAX_NUCL,sizeof(char));
    all_seq[j].name = (string_t) calloc(SIZE_MAX_NAME,sizeof(char));
    all_seq[j].cofold = new_vector (nb_seq);
    for (k=0; k<nb_seq; k++) all_seq[j].cofold[k] = 1;
    all_seq[j].N = -1;
    // sequence name + skip comments
    skip = 1; while ((skip) && (c != EOF)){
      while ((c != '$') && (c != '>') && (c != EOF)) c=fgetc(f);
      fgets((all_seq[j].name),SIZE_MAX_NAME,f);
      if (c == '$') c='*'; if (c == '>') skip=0;}
    // erase blancs at the ends
    n = strlen(all_seq[j].name); all_seq[j].name[n-1]='\0';
    for (k=0; k<n; k++) {c = all_seq[j].name[k];
    if ((c=='\\') || (c=='/') || (c==' ')) all_seq[j].name[k] = '_';}
    k=2; while (all_seq[j].name[n-k]=='_') {all_seq[j].name[n-k]='\0'; k++;}
    k=0; while (all_seq[j].name[k]=='_') {all_seq[j].name++; k++;}
    if (VERBOSE) printf("\n  reading: [%2i]", j+1);
    all_seq[j].discard = 0;
    // sequence: nucl = nucl[0..n-1]
    i=0; while (((c = fgetc(f)) != '>') && (c != '$') && (c != EOF))
      {if (is (c,base))	(all_seq[j].nucl)[i++]=(char)tolower((int)c);
      else if ((c=='t')||(c=='T')) (all_seq[j].nucl)[i++]='u';}
    (all_seq[j].nucl)[i]='\0';
    // a, u, c, g
    aa = 0; uu = 0; cc = 0; gg = 0;
    all_seq[j].length = strlen(all_seq[j].nucl);
    for (i=0; i<all_seq[j].length; i++) {
      if (all_seq[j].nucl[i] == 'a') aa++;
      if (all_seq[j].nucl[i] == 'u') uu++;
      if (all_seq[j].nucl[i] == 'c') cc++;
      if (all_seq[j].nucl[i] == 'g') gg++;}
    //    printf("\t[a %3i u %3i c %3i g %3i]", aa, uu, cc, gg);
    au = (aa+uu) * 100 / all_seq[j].length;
    all_seq[j].cg = (cc+gg) * 100 / all_seq[j].length;
    if (VERBOSE) {
      printf (" [l %3i]", all_seq[j].length);
      printf (" [cg %2i %c] %s", all_seq[j].cg, '%', all_seq[j].name);
      fflush(stdout);
    }
  }
  fclose(f);

  // no sequence
  if (nb_seq==0)
    {if (VERBOSE) printf("  no sequence was read by CARNAC\n\n"); exit(0);}

  // allocate memory: R, finalR & remain
  R = (stem_list_t*) malloc (nb_seq * sizeof(stem_list_t));
  for (i=0; i<nb_seq; i++) R[i] = new_stem_list (0);
  finalR = (stem_list_t*) malloc (nb_seq * sizeof(stem_list_t));
  for (i=0; i<nb_seq; i++) finalR[i] = new_stem_list (0);
  remain = (stem_list_t*) malloc (nb_seq * sizeof(stem_list_t));
  for (i=0; i<nb_seq; i++) remain[i] = new_stem_list (0);

  return all_seq;
}

int simil (sequence_list_t all_seq, int ii, int jj) {
  matrix2_t m; int i, j, n0, n1, ans; char a, b;
  n0 = all_seq[ii].length;
  n1 = all_seq[jj].length;
  m = alloc_matrix2 (n0, n1, 0);
  // match = +1
  // subst = -1
  // indel = -2
  for (i=0; i<n0; i++) {
    for (j=0; j<n1; j++) {
      a = all_seq[ii].nucl[i];
      b = all_seq[jj].nucl[j];
      m[i][j] = m[i-1][j-1] + ((a)==(b) ? (1):(-1));
      m[i][j] = MAX (m[i][j], m[i-1][j] - 2);
      m[i][j] = MAX (m[i][j], m[i][j-1] - 2);}}
  // global
  ans = m[n0-1][n1-1];
  // neglecting ends
  for (i=0; i<n0; i++) ans = MAX (ans, m[i][n1-1]);
  for (j=0; j<n1; j++) ans = MAX (ans, m[n0-1][j]);
  ans = 100 * ans / MIN (n0, n1);
  if (DEBUG) {FILE *out; out = fopen ("output/simil.out", "a");
  fprintf (out, "\n[%2i:%2i]:%3i %c", ii+1, jj+1, ans, '%'); fclose (out);}
  free_matrix2 (m, n0); return ans;}

sequence_list_t reduce_set (sequence_list_t all_seq) {

  extern matrix2_t similarity;
  extern int nb_seq;
  int i, j, clock, new_clk, old_clk, size;
  sequence_list_t new_all_seq;

  if (VERBOSE && !CLEAN) line;
  if (VERBOSE && CLEAN)
    printf("\n\n  discard sequences too similar (> %i%c) - ", SIMIL_THD, '%');
  // init clock
  clock = 0; new_clk = 0; old_clk = -1;
  size = nb_seq * (nb_seq-1) / 2;
  if (VERBOSE && DYNAM && CLEAN) printf("     ");

  for (i=0; i<nb_seq; i++) {
    for (j=i+1; j<nb_seq; j++) {

      // ----- clock --------------------------------------------
      if (VERBOSE && DYNAM && CLEAN) {
	new_clk =  clock * 100 / size;
	if (new_clk-old_clk) {old_clk = new_clk; printf("\b\b\b\b\b");
	printf("%3i %c", new_clk,'%'); fflush(stdout);}}
      // ----- clock --------------------------------------------

      similarity [i][j] = simil(all_seq, i,j);

      clock++;
      if (
	  (CLEAN) &&
	  (similarity [i][j] >= SIMIL_THD) &&
	  (!all_seq[i].discard) &&
	  (!all_seq[j].discard)
	  ) {
	// discard the longest one
	if (all_seq[i].length > all_seq[j].length) {
	  all_seq[i].discard = 1;
	  if (VERBOSE) printf ("\b\b\b\b\b[%2i]     ", i+1);
	  discd [i] = 1;}
	else {
	  all_seq[j].discard = 1;
	  if (VERBOSE) printf ("\b\b\b\b\b[%2i]     ", j+1);
	  discd [j] = 1;}
	discarded++;
      }
    }
  }

  // print matrix of similarities
  if (DEBUG) {
    FILE *out;
    out = fopen ("output/simil_matrix.out", "w");
    fprintf (out, "\n    ");
    for (j=0; j<nb_seq; j++) fprintf (out, "%2i ", j+1); fprintf (out, "\n\n");
    for (i=0; i<nb_seq; i++) {fprintf (out, "%2i: ", i+1);
    for (j=0; j<nb_seq; j++) {
      if ((i==j) || (similarity [i][j] == -1)) fprintf (out, " - ");
      else if (similarity [i][j] == 100) fprintf (out, " M ");
      else if (similarity [i][j]) fprintf (out, "%2i ", similarity [i][j]);
      else fprintf (out, "   ");}
    fprintf (out, "\n");}
    fclose (out);}

  // display how many sequences are discarded
  if (VERBOSE && DYNAM && CLEAN) printf("\b\b\b\b\b100 %c",'%');
  if (VERBOSE && CLEAN) {
    printf ("\b\b\b\b\b ");
    if (discarded) printf ("deleted");
    else printf ("none");
    printf ("\n");}

  // read sequences
  nb_seq -= discarded;
  new_all_seq = (sequence_list_t) malloc (nb_seq * sizeof(sequence_t));
  if (VERBOSE && discarded) printf("\n  new sequence set:\n");
  j=0; for (i=0; i<nb_seq+discarded; i++) {
    if (!all_seq[i].discard) {
      if (VERBOSE && discarded) printf("\n  [%2i] %s", j+1, all_seq[i].name);
      new_all_seq[j] = all_seq[i]; j++;
    }
  }

  free (all_seq);
  return new_all_seq;
}

void get_true_structure () {

  extern string_t refname;
  extern int nb_seq;
  extern stem_list_t *all_stems;
  extern sequence_list_t all_seq;
  matrix2_t bp;
  string_t base = "AUCGaucg";
  vector_t length;
  int correct_bp_in;
  int ave_correct_bp_in;
  int i, j, k, l, x, y, m, n;
  int a,b,c,d,aa,bb,cc,dd;
  int current_shift, flag;
  vector_t shift; // skip any base != 'aucg'
  int bp_cpl[5000][2];
  string_t ctcatname;
  char tmp;
  FILE *f, *compare;
  stem_t P;

  ctcatname = cat (new_string(refname), new_string (".cct"));
  if ((f=fopen(ctcatname,"r")) != NULL) {

	 COMPARE = 1;
	 compare = fopen("output/compare.out","w");
	 fprintf (compare, "\n");
	 fprintf (compare, "        name       -");
	 fprintf (compare, " lgth -");
	 fprintf (compare, " input bp -");
	 fprintf (compare, " miss");
	 fprintf (compare, "\n\n");

	 if (VERBOSE) printf ("  reading structure file: %s.cct\n\n", refname);

	 // allocate memory
	 true_stems = (stem_list_t*) malloc (nb_seq * sizeof(stem_list_t));
	 true_bp = new_vector (nb_seq);
	 correct_stems_in = new_vector (nb_seq);
	 for (x=0; x<nb_seq; x++)
	   true_stems[x] = new_stem_list (NB_MAX_TRUE_STEMS);
	 length = new_vector (nb_seq);

	 ave_correct_bp_in = 0;
	 ave_correct_stems_in = 0;
	 ave_true_bp = 0;
	 y = -1;

	 // parse 'ct' files
	 for (x=0; x<nb_seq; x++) {

	   y++;
	   // scan first line (length)
	   fscanf (f, "%i", &length[x]); while (fgetc(f)!='\n');
	   // skip discarded sequence
	   if (CLEAN)
	     while (discd [y]) {y++; for (m=0; m<length[x]; m++)
	       fscanf (f, "%*i %*c %*i %*i %*i %*i");
	     fscanf (f, "%i", &length[x]); while (fgetc(f)!='\n');}

	   // init shift
	   current_shift = 0; shift = new_vector (length[x]); shift[0]=0;
	   // init stats
	   true_bp[x] = 0; correct_stems_in[x] = 0;

	   for (m=0; m<length[x]; m++) {
	     // store pairings in bp_cpl
	     fscanf(f, "%i %c %*i %*i %i %*i",
		    &(bp_cpl[m][0]), &tmp, &(bp_cpl[m][1]));
	     if (bp_cpl[m][1]) true_bp[x]++;
	     if (!is (tmp, base)) current_shift++;
	     shift[m]=current_shift;
	   }

	   // create stem list
	   //      printf ("[seq %3i: #'N'= %2i] ", x+1, current_shift);
	   //      for (m=0; m<length[x]; m++) print_ (shift[m]); nnn;
	   n = 0; for(m=0; m<length[x]; m++) {
	     if (bp_cpl[m][1] > bp_cpl[m][0]) {
	       i = bp_cpl[m][0]-1-shift[m];
	       l = bp_cpl[m][1]-1-shift[m];
	       m++;
	       while ((bp_cpl[m][1]) &&
		      (bp_cpl[m][1] == bp_cpl[m-1][1] - 1) &&
		      (bp_cpl[m][0] == bp_cpl[m-1][0] + 1)) m++;
	       m--;
	       j = bp_cpl[m][0]-1-shift[m];
	       k = bp_cpl[m][1]-1-shift[m];
	       true_stems[x][n] = new_stem (x, n, 0, i, j, k, l, 0); n++;
	     }
	   }

	   free (shift);

	   // compare stem lists
	   //       a ..... b      c ..... d       true stem
	   //   aa ..... bb          cc ..... dd   potential stem

	   bp = alloc_matrix2 (length[x], length[x], 0);
	   n = stem_list_length (all_stems[x]);
	   m = stem_list_length (true_stems[x]);

	   correct_bp_in = 0;
	   
	   for (j=0; j<n; j++) {
	     for (i=0; i<m; i++) {
	       // true  [i]
	       a = true_stems[x][i]->s[1]+1;
	       b = true_stems[x][i]->s[2]+1;
	       c = true_stems[x][i]->s[3]+1;
	       d = true_stems[x][i]->s[4]+1;
	       // potential  [j]
	       aa = all_stems[x][j]->s[1]+1;
	       bb = all_stems[x][j]->s[2]+1;
	       cc = all_stems[x][j]->s[3]+1;
	       dd = all_stems[x][j]->s[4]+1;

	       P = all_stems[x][j];

	       if (((a<=aa) && (aa<=b)) || ((aa<=a) && (a<=bb))) {
		 if (a+d == aa+dd) {
		   true_stems[x][i] -> true_match +=10;
		   P -> true_match +=10;
		   if ((P -> true_match >= 10) &&
		       (P -> true_match < 20)) {
		     // how many common bp & stems
		     if (!P->substem) {
		       correct_stems_in[x]++;
		       for (k=0; k<=bb-aa; k++) for (l=0; l<=b-a; l++)
			 if ((aa+k == a+l) && (dd-k == d-l) &&
			     (bp [a+l][d-l] < 10)) {
			   bp [a+l][d-l] = 10;
			   correct_bp_in ++;
			 }
		     }
		   }
		 }
		 else if (((c<cc) && (d>=cc)) || ((c>=cc) && (c<=dd)))
		   {true_stems[x][i]->true_match ++;
		   all_stems[x][j]->true_match ++;
		   }
	       }
	     }
	   }

	   // print bp comparison with true structure
	   free_matrix2 (bp, length[x]);
	   true_bp[x] /= 2;
	   ave_correct_bp_in += correct_bp_in;
	   ave_correct_stems_in += correct_stems_in[x];
	   ave_true_bp += true_bp[x];
	   flag = (strlen(all_seq[x].name)>=12);
	   if (flag) {tmp=all_seq[x].name[12]; all_seq[x].name[12]='\0';}
	   fprintf (compare, "[%2i] %s ", x+1, all_seq[x].name);
	   if (flag) all_seq[x].name[12]=tmp;
	   fprintf (compare, " - %4i", length[x]);
	   fprintf (compare, " - %3i/%3i ", correct_bp_in, true_bp[x]);
	   if (true_bp[x])
	     fprintf (compare, " - %2i %c\n",
		      100 - 100 * correct_bp_in / true_bp[x], '%');
	   print_reference_stems (all_seq, x, true_stems[x]);

	 }

	 ave_correct_bp_in = RND_FRAC (ave_correct_bp_in, nb_seq);
	 ave_correct_stems_in = RND_FRAC (ave_correct_stems_in, nb_seq);
	 ave_true_bp = RND_FRAC (ave_true_bp, nb_seq);
	 fprintf (compare, "\n");
	 fprintf (compare, "       average           ");
	 fprintf (compare, " - %3i/%3i ", ave_correct_bp_in, ave_true_bp);
	 if (ave_true_bp)
	   fprintf (compare, " - %2i %c\n",
		    100 - 100 * ave_correct_bp_in / ave_true_bp, '%');
	 fprintf (compare, "\n");
	 fclose (compare);
	 fclose (f);

  }
}

void final_compare () {

  extern int nb_seq;
  extern stem_list_t *all_stems, *R, *finalR;
  extern sequence_list_t all_seq;
  int i, j, k, l, m, n, nn, x;
  int a,b,c,d,aa,bb,cc,dd;
  int beg, end, len;
  matrix2_t bp;
  int length, flag;
  char tmp = '\0';
  FILE *compare;
  int ave_n, shifts, bp_out, ave_bp_out;
  int correct_bp_out, ave_correct_bp_out;
  int correct_stems_out, ave_correct_stems_out;
  int correct_stems_folded;

  compare = fopen("output/compare.out","a");
  fprintf (compare, "\n");
  fprintf (compare, "        name       -");
  fprintf (compare, " in => fo => sel (sh)");
  fprintf (compare, " total -");
  fprintf (compare, " output bp -");
  fprintf (compare, " correct -");
  fprintf (compare, "  miss");
  fprintf (compare, "\n\n");
  fclose (compare);

  // compare results with true structure
  // compare stem lists
  //       a ..... b      c ..... d       true stem
  //   aa ..... bb          cc ..... dd   folded stem

  ave_correct_stems_out = 0;
  ave_correct_bp_out = 0;
  ave_bp_out = 0;
  ave_n = 0;

  for (x=0; x<nb_seq; x++) {

    sort_stems_by_begin (all_stems[x]);
    print_true_stems (all_seq, x, all_stems[x]);

    length = strlen (all_seq[x].nucl);
    bp = alloc_matrix2 (length, length, 0);
    m = stem_list_length (true_stems[x]);
    n = stem_list_length (finalR[x]);
    nn = stem_list_length (R[x]);

    ave_n += n;
    shifts = 0;
    bp_out = 0;
    correct_stems_out = 0;
    for (j=0; j<n; j++) {
      beg = finalR[x][j]->s[1];
      end = finalR[x][j]->s[4];
      len = finalR[x][j]->s[2] - finalR[x][j]->s[1] + 1;
      for (k=0; k<len; k++)
	if (bp[beg+k][end-k] == 0) {bp_out++; bp[beg+k][end-k]=1;}
      if (finalR[x][j] -> true_match >= 10) correct_stems_out++;
      else if (finalR[x][j] -> true_match > 0) shifts++;}
    ave_correct_stems_out += correct_stems_out;
    ave_bp_out += bp_out;

    correct_stems_folded = 0;
    for (j=0; j<nn; j++)
      if ((!R[x][j]->substem) &&
	  (R[x][j] -> true_match >= 10))
	correct_stems_folded++;

    correct_bp_out = 0;
    for (j=0; j<n; j++) {
      for (i=0; i<m; i++) {

	// potential  [j]
	aa = finalR[x][j]->s[1]+1;
	bb = finalR[x][j]->s[2]+1;
	cc = finalR[x][j]->s[3]+1;
	dd = finalR[x][j]->s[4]+1;
	// true  [i]
	a = true_stems[x][i]->s[1]+1;
	b = true_stems[x][i]->s[2]+1;
	c = true_stems[x][i]->s[3]+1;
	d = true_stems[x][i]->s[4]+1;

	if ((((a<=aa) && (aa<=b)) ||
	     ((aa<=a) && (a<=bb))) &&
	    (a+d == aa+dd)) {
	  for (k=0; k<=bb-aa; k++) for (l=0; l<=b-a; l++)
	    if ((aa+k == a+l) && (dd-k == d-l) &&
		(bp [a+l][d-l] < 10)) {
	      bp [a+l][d-l] = 10;
	      correct_bp_out++;
	    }
	}
      }
    }

    // print bp comparison with true structure
    free_matrix2 (bp, length);
    ave_correct_bp_out += correct_bp_out;
    compare = fopen("output/compare.out","a");
    flag = (strlen(all_seq[x].name)>=12);
    if (flag) {tmp=all_seq[x].name[12]; all_seq[x].name[12]='\0';}
    fprintf (compare, "[%2i] %s ", x+1, all_seq[x].name);
    if (flag) all_seq[x].name[12]=tmp;
    fprintf (compare, " -%3i =>%3i =>%3i",
	     correct_stems_in[x],
	     correct_stems_folded,
	     correct_stems_out);
    if (shifts) fprintf (compare, "   %i   ", shifts);
    else fprintf (compare, "       ");
    fprintf (compare, " %2i", n);
    if (n == correct_stems_out + shifts) fprintf (compare, " *");
    else fprintf (compare, "  ");
    fprintf (compare, " -  %3i/%3i ", correct_bp_out, bp_out);
    if (bp_out)
      fprintf (compare, " -  %3i %c ", 100 * correct_bp_out / bp_out, '%');
    if (true_bp[x])
      fprintf (compare, " - %3i %c\n",
	       100 - 100 * correct_bp_out / true_bp[x], '%');
    fclose (compare);
  }

  compare = fopen("output/compare.out","a");
  ave_correct_stems_out = RND_FRAC (ave_correct_stems_out, nb_seq);
  ave_correct_bp_out = RND_FRAC (ave_correct_bp_out, nb_seq);
  ave_bp_out = RND_FRAC (ave_bp_out, nb_seq);
  ave_n = RND_FRAC (ave_n, nb_seq);
  fprintf (compare, "\n");
  fprintf (compare, "       average     -");
  fprintf (compare, "%3i =======>", ave_correct_stems_in);
  fprintf (compare, "%3i       ", ave_correct_stems_out);
  fprintf (compare, "%3i  ", ave_n);
  fprintf (compare, " -  %3i/%3i ", ave_correct_bp_out, ave_bp_out);
  if (ave_bp_out)
    fprintf (compare, " -  %3i %c ", 100 * ave_correct_bp_out / ave_bp_out, '%');
  if (ave_true_bp)
    fprintf (compare, " - %3i %c\n", 100 - 100 * ave_correct_bp_out / ave_true_bp, '%');
  fprintf (compare, "\n");
  fclose (compare);
}



void tournament (int i, int j, stem_list_t *IN, stem_list_t *all0, stem_list_t *all1) {
  extern sequence_list_t all_seq, seq;
  int ii, jj, n0, n1;
  if (VERBOSE) printf("[%2i:%2i] ", i+1, j+1); fflush(stdout);
  seq[0] = all_seq[i]; seq[1] = all_seq[j];
  (*all0) = IN[i]; (*all1) = IN[j];
  n0 = stem_list_length (*all0);
  n1 = stem_list_length (*all1);
  for (ii=0; ii<n0; ii++) (*all0)[ii]->lonely=1;
  for (jj=0; jj<n1; jj++) (*all1)[jj]->lonely=1;}
