

/*----------------------------------------------*/
/* graph.h                                      */
/*----------------------------------------------*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "types.h"
#include "parse.h"
#include "string.h"

#ifndef RATIO_NB_MAX_G_NODES
#define RATIO_NB_MAX_G_NODES           200     // 200
#define RATIO_NB_MAX_COMPOUNDS         50      // 50
#define RATIO_NB_MAX_LINKS_BY_G_NODE   8       // 5
#define RATIO_NB_NODES_BY_COMPOUND     5       // 3
// SIZE = RATIO x nb_seq

typedef struct _g_node_t {   // a node of the graph
  int seq;                   // which sequence
  int rank;		     // rank in the original node list
  int visited;		     // bool: yet visited ?
  int compound;		     // which compound
  int true_class;	     // bool: is it a true class ?
  stem_list_t in_stem_list;  // content = stems
  int arity_out;	     // links to cofolded stems
  int arity_id;              // links to identical stems
  struct _g_node_t **out_g_node_list;
  struct _g_node_t **id_g_node_list;
} _g_node_t, *g_node_t;
typedef g_node_t *g_node_list_t;

typedef struct _compound_t {
  struct _g_node_t ** list;
  int size;
  int corrected_size;
  int sq;
  int co;
  int id;
  int me;
  int sig;   // is significant ? (0..100)
  int rel;   // is reliable ?    (0..100)
  int index;
} _compound_t, *compound_t;
typedef compound_t *compound_list_t;

#endif

g_node_list_t new_g_node_list (int size);
g_node_t new_g_node (int seq, int rank);
int g_node_list_length (g_node_list_t gc);
void add_to_out_g_node_list (g_node_t G, g_node_t out);
void add_to_id_g_node_list (g_node_t G, g_node_t out);
void free_g_node_list (g_node_list_t G, int size);

compound_list_t new_compound_list (int size);
compound_t new_compound (int size);
int add_g_node_to_compound (compound_t C, g_node_t G);
void delete_g_node (g_node_list_t S, int i, g_node_t G);
void free_compound_list (compound_list_t CL, int size);

void sort_list_compound(compound_t C);
void sort_out_g_node_list (g_node_t G);
void sort_compounds_list_by_sig (compound_list_t C, int length);
void sort_compounds_list_by_rel (compound_list_t C, int length);
void sort_compounds_list_by_index (compound_list_t C, int length);

int belongs2g_node_list (g_node_t G, g_node_list_t Glist);
int rank (g_node_list_t S, g_node_t G);

void print_class_content (g_node_list_t g_node, int i);
void print_class_connect (g_node_list_t g_node, int i);
void print_connex_compounds (compound_list_t compound_list, int i);
void print_graph_infos (compound_t C);
void close_graph_infos ();

void compute_stats (compound_t C);
void generate_dot_file (compound_list_t compound_list, int i);

