
// parse.c

#include "parse.h"
#include <unistd.h>
//++pac 
#ifndef Plan9
#include <malloc.h>
#endif
#include <string.h>

#define LINE_SIZE 2048

void skip_blank (char *str, int *i) {
  int j; int k = *i;
  do {j = *i;
  if ((str[*i] == '#') || (str[*i] == '/') || (str[*i] == '$'))
    while (str[*i] != '\n') (*i)++;
  while (str[*i] == ' ')  (*i)++;
  while (str[*i] == ',')  (*i)++;
  while (str[*i] == ';')  (*i)++;
  while (str[*i] == '\t') (*i)++;
  while (str[*i] == '\n') (*i)++;
  while (str[*i] < 31)    (*i)++;}
  while (*i != j);
  if (k != *i) (*i)--;}

char get_boolean (char *str)
{switch (str[0]) {case 'Y': case 'y': case 'o':
 case 'O': case '1': return 1; default: return 0;}}

void parse_config_file (char *fileName) {
  FILE *file; char  Line[LINE_SIZE];
  int i; int tmp_i; float tmp_f;
  if ((file=fopen(fileName,"r")) == NULL)
    {fprintf(stderr, "carnac: unknown file %s\n", fileName); exit(0);}
  while (fgets (Line, LINE_SIZE, file) != NULL) {
    i = 0; skip_blank (Line, &i);
    if (1 == sscanf (Line+i, "VERBOSE = %i", &tmp_i))
      VERBOSE = tmp_i;
    else if (1 == sscanf (Line+i, "AP_THD = %i", &tmp_i))
      AP_THD = tmp_i;
    else if (1 == sscanf (Line+i, "SUB = %i", &tmp_i))
      SUB = tmp_i;
    else if (1 == sscanf (Line+i, "SIMIL_THD = %i", &tmp_i))
      SIMIL_THD = tmp_i;
    else if (1 == sscanf (Line+i, "SIZE_HAIRPIN = %i", &tmp_i))
      SIZE_HAIRPIN = tmp_i;
    else if (1 == sscanf (Line+i, "SCALING = %f", &tmp_f))
      SCALING = tmp_f;
    else if (1 == sscanf (Line+i, "CORRECT_THD = %i", &tmp_i))
      CORRECT_THD = tmp_i;
    else if (1 == sscanf (Line+i, "INI_THD = %i", &tmp_i))
      INI_THD = tmp_i;
    else if (1 == sscanf (Line+i, "DIST_1 = %i", &tmp_i))
      DIST_1 = tmp_i;
    else if (1 == sscanf (Line+i, "THD_1 = %i", &tmp_i))
      THD_1 = tmp_i;
    else if (1 == sscanf (Line+i, "DIST_2 = %i", &tmp_i))
      DIST_2 = tmp_i;
    else if (1 == sscanf (Line+i, "THD_2 = %i", &tmp_i))
      THD_2 =tmp_i;
    else if (1 == sscanf (Line+i, "SINGLE_HP = %i", &tmp_i))
      SINGLE_HP = tmp_i;
    else if (1 == sscanf (Line+i, "SIZE_MAX_HP = %i", &tmp_i))
      SIZE_MAX_HP = tmp_i;
    else if (1 == sscanf (Line+i, "THD_HP = %i", &tmp_i))
      THD_HP = tmp_i;
    else if (1 == sscanf (Line+i, "COVAR = %i", &tmp_i))
      COVAR = tmp_i;
    else if (1 == sscanf (Line+i, "FLT = %i", &tmp_i))
      FLT = tmp_i;
    else if (1 == sscanf (Line+i, "LARGE = %i", &tmp_i))
      LARGE = tmp_i;}
  INI_THD = (int) (SCALING * (float) INI_THD);
  THD_1 =   (int) (SCALING * (float) THD_1  );
  THD_2 =   (int) (SCALING * (float) THD_2  );
  THD_HP =  (int) (SCALING * (float) THD_HP );
  fclose (file);}

void parse_command_line (int argc, char **argv) {
  int c;
  extern char* optarg;
  extern char* infile_arg;
  extern char* infile_seq;
  extern char* infile_stem;
  extern char* infile_forced;
  if (argc < 2) {
    printf ("usage : %s [-p <filename.in>] [-e -f <filename.rnaml>] [-scodigytl] <filename.fasta>\n", argv[0]);
    exit (0);}
  infile_arg = "./params/default.in";
  infile_stem = NULL;
  INTERFACE = 0;
  VERBOSE = 1;
  DYNAM = 1;
  OUTPUT = 0;
  GRAPH = 0;
  COMPARE = 0;
  CLEAN = 0;
  DEBUG = 0;
  TRY = 0;
  LATEX = 0;
  while ((c = getopt(argc,argv,"p:e:f:scodigytl")) != EOF) {
    if (c =='p') {
      infile_arg = (char*) malloc (strlen(optarg)+1 * sizeof(char));
      strcpy (infile_arg, optarg);}
    else if (c =='e') {
      infile_stem = (char*) malloc (strlen(optarg)+1 * sizeof(char));
      strcpy (infile_stem, optarg);}
    else if (c =='f') {
      infile_forced = (char*) malloc (strlen(optarg)+1 * sizeof(char));
      strcpy (infile_forced, optarg);}
    else if (c =='s') VERBOSE = 0;  // silent mode
    else if (c =='y') DYNAM = 0;    // undisplay percentages
    else if (c =='o') OUTPUT = 1;   // pairwise output files
    else if (c =='g') GRAPH = 1;    // create graph files '.dot'
    else if (c =='c') CLEAN = 1;    // discard sequences too similar
    else if (c =='d') DEBUG = 1;    // files 'check_*.out'
    else if (c =='t') TRY = 1;      // private tests
    else if (c =='l') LATEX = 1;    // generates latex file
    else if (c =='i') {VERBOSE = 0; INTERFACE = 1;}
    else exit(0);}
  infile_seq = (char*) malloc (strlen(argv[argc-1])+1 * sizeof(char));
  strcpy (infile_seq, argv[argc-1]);}
